/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import org.aopalliance.intercept.MethodInvocation;
import org.springframework.aop.IntroductionInterceptor;

/**
 * Intercepts finder DAO calls and calls execute*Finder methods on dao.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
final class FinderIntroductionInterceptor implements IntroductionInterceptor {

  /** {@inheritDoc} */
  public boolean implementsInterface(final Class intf) {
    return intf.isInterface() && FinderExecutor.class.isAssignableFrom(intf);
  }

  /** {@inheritDoc} */
  @Override
  public Object invoke(final MethodInvocation invocation) throws Throwable {
    final FinderExecutor dao = (FinderExecutor) invocation.getThis();
    final Method meth = invocation.getMethod();
    if (Utils.isFinder(meth)) {
      final Object[] arguments = invocation.getArguments();
      if (Utils.isSingleFind(invocation.getMethod())) {
        final Map<String, ?> namedArgs = getNamedParams(meth, arguments);
        if (namedArgs == null || namedArgs.isEmpty()) {
          return dao.executeSingleFinder(meth, arguments);
        } else {
          return dao.executeSingleFinder(meth, namedArgs);
        }
      } else {
        // TODO: cache indices between invocations?
        final Integer limitIndex = Utils.getLimitParameterIndex(meth);
        final Integer offsetIndex = Utils.getOffsetParameterIndex(meth);

        int end = arguments.length;
        // This works because we said that limit/offset can only be last two parameters.
        if (limitIndex != null) {
          end--;
        }
        if (offsetIndex != null && !offsetIndex.equals(limitIndex)) {
          end--;
        }
        final Integer limit = limitIndex == null ? null : (Integer) arguments[limitIndex];
        final Integer offset = offsetIndex == null ? null : (Integer) arguments[offsetIndex];
        final Object[] queryArgs = end == arguments.length ? arguments : Arrays.copyOfRange(arguments, 0, end);
        final Map<String, ?> namedArgs = getNamedParams(meth, queryArgs);
        if (namedArgs == null || namedArgs.isEmpty()) {
          return dao.executeListFinder(meth, queryArgs, limit, offset);
        } else {
          return dao.executeListFinder(meth, namedArgs, limit, offset);
        }
      }
    } else {
      return invocation.proceed();
    }
  }

  private static Map<String, ?> getNamedParams(final Method meth, final Object[] queryArgs) {
    final Map<String, Object> result = new HashMap<String, Object>();
    if (queryArgs.length > 0) {
      for (int i = 0; i < queryArgs.length; i++) {
        Named n = Utils.getNamed(meth.getParameterAnnotations()[i]);
        if (n == null) {
          return null;
        } else {
          result.put(n.value(), queryArgs[i]);
        }
      }
    }
    return result;
  }
}
