/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao.hibernate;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import net.sf.autodao.Dao;
import net.sf.autodao.Named;
import net.sf.autodao.PersistentEntity;
import net.sf.autodao.SpecificationViolationException;
import net.sf.autodao.Utils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.SessionFactory;
import org.hibernate.engine.SessionFactoryImplementor;
import org.springframework.aop.framework.ProxyFactoryBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Required;
import org.springframework.util.Assert;

/**
 * Validating proxy factory.
 * <p/>NOT FOR PUBLIC USE.
 */
public final class HibernateDaoProxyFactoryBean extends ProxyFactoryBean implements InitializingBean {

  /** Log. Just log. */
  private static final Log LOG = LogFactory.getLog(HibernateDaoProxyFactoryBean.class);

  /** Hibernate session factory. Required to extract class metadata. */
  private SessionFactory sessionFactory;

  /** Entity type. */
  private Class<? extends PersistentEntity<?>> type;

  private static void validateParams(final Method meth) {
    final boolean isSingleFind = Utils.isSingleFind(meth);
    final Integer limitIndex = Utils.getLimitParameterIndex(meth);
    final Integer offsetIndex = Utils.getOffsetParameterIndex(meth);
    final Set<String> names = new HashSet<String>();
    Boolean hasNamed = null;
    for (int i = 0; i < meth.getParameterAnnotations().length; i++) {
      final Class<?> paramType = meth.getParameterTypes()[i];
      final Named named = Utils.getNamed(meth.getParameterAnnotations()[i]);
      final boolean isOffset = Utils.getOffset(meth.getParameterAnnotations()[i]) != null;
      final boolean isLimit = Utils.getLimit(meth.getParameterAnnotations()[i]) != null;
      if (!isLimit && !isOffset) {
        if (hasNamed == null) {
          hasNamed = named != null;
        } else {
          if (hasNamed && named == null || !hasNamed && named != null) {
            throw new SpecificationViolationException("You're not allowed to mix @Named and indexed params", meth);
          }
        }
      } else {
        if (named != null) {
          throw new SpecificationViolationException("@Named cannot be present on @Limit/@Offset param", meth);
        }
      }
      if (named != null && !names.add(named.value())) {
        throw new SpecificationViolationException("Duplicate named parameters '" + named.value() + "'", meth);
      }
      if (isSingleFind) {
        if (isLimit || isOffset) {
          throw new SpecificationViolationException("Neither @Limit nor @Offset is allowed on single-object finders",
              meth);
        }
      } else {
        if ((isLimit || isOffset) && !Integer.class.isAssignableFrom(paramType)
            && !Integer.TYPE.isAssignableFrom(paramType)) {
          throw new SpecificationViolationException(
              "@Limit and @Offset argument type must be either int or java.lang.Integer", meth);
        }
        if (isLimit && limitIndex != i || isOffset && offsetIndex != i) {
          throw new SpecificationViolationException("@Limit and @Offset can appear only on one parameter", meth);
        }
        if (!isLimit && !isOffset && limitIndex != null && i > limitIndex || offsetIndex != null && i > offsetIndex) {
          throw new SpecificationViolationException(
              "There cannot be any parameters after @Limit and @Offset parameters", meth);
        }
      }
    }
  }

  /** {@inheritDoc} */
  public void afterPropertiesSet() {
    // Do not trust anyone. Especially yourself.
    Assert.notNull(this.type, "No entity type given");
    Assert.isAssignable(PersistentEntity.class, this.type, "Invalid entity type");
    Assert.isTrue(super.getProxiedInterfaces().length == 1, "Expected only single proxy interface");

    final Class<?> daoInterface = super.getProxiedInterfaces()[0];
    Assert.isAssignable(Dao.class, daoInterface, "Invalid DAO interface");
    this.validate(daoInterface);
  }

  public Class<? extends PersistentEntity<?>> getType() {
    return this.type;
  }

  @Required
  public void setSessionFactory(final SessionFactory sessionFactory) {
    this.sessionFactory = sessionFactory;
  }

  @Required
  public void setType(final Class<? extends PersistentEntity<?>> type) {
    this.type = type;
  }

  /**
   * Validates given DAO interface.
   * @param daoInterface dao interface.
   */
  private void validate(final Class<?> daoInterface) {
    SessionFactoryImplementor sfi = null;
    if (this.sessionFactory == null) {
      HibernateDaoProxyFactoryBean.LOG.warn("No session factory given. Some validations were disabled for DAO "
          + daoInterface.getName());
    } else {
      Assert.notNull(this.sessionFactory.getClassMetadata(this.type), this.type.getName()
          + " is not a persistent class (it is not registered in SessionFactory)");
      // Have a better way? Tell me.
      try {
        sfi = (SessionFactoryImplementor) this.sessionFactory;
      } catch (final ClassCastException e) {
        HibernateDaoProxyFactoryBean.LOG.warn("Could not cast session factory " + this.sessionFactory + " to "
            + SessionFactoryImplementor.class.getName() + ". Some validations were disabled for DAO "
            + daoInterface.getName());
      }
    }

    final Set<String> seen = new HashSet<String>();
    for (final Method m : daoInterface.getMethods()) {
      if (Utils.isFinder(m)) {
        Assert.isTrue(seen.add(m.getName()), "AutoDAO doesn't support method overloading. Offending method: " + m);
        if (sfi != null) {
          final String queryName = HibernateUtils.getNamedQueryName(this.sessionFactory, this.type, m);
          Assert.isTrue(sfi.getNamedQuery(queryName) != null || sfi.getNamedSQLQuery(queryName) != null,
              "No named query '" + queryName + "' found. Method " + m + " cannot work.");
        }
        HibernateDaoProxyFactoryBean.validateParams(m);
      } else {
        if (Dao.class.isAssignableFrom(daoInterface)) {
          Assert.isTrue(Arrays.asList(Dao.class.getMethods()).contains(m), "Method " + m
              + " is neigher a finder method nor is implemented in AutoDAO");
        } else {
          throw new IllegalArgumentException("Method " + m + " is not a finder method");
        }
      }
    }
  }
}
